<?php

namespace Drupal\media_files_handler\Service;

use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\file\Entity\File;

/**
 * Class HandlerService.
 */
class HandlerService {

  /**
   * The default delete action.
   *
   * Todo: Expose as Configuration.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  private $deleteAction = 'temporary';

  /**
   * Drupal file_system service.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected $fileSystem;

  /**
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The media storage.
   *
   * @var \Drupal\Core\Entity\ContentEntityStorageInterface
   */
  protected $mediaStorage;

  /**
   * The file storage.
   *
   * @var \Drupal\Core\Entity\ContentEntityStorageInterface
   */
  protected $fileStorage;

  /**
   * Constructs a new HandlerService object.
   *
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   *   The file system.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function __construct(FileSystemInterface $file_system, EntityTypeManagerInterface $entity_type_manager) {
    $this->fileSystem = $file_system;
    $this->entityTypeManager = $entity_type_manager;
    $this->mediaStorage = $entity_type_manager->getStorage('media');
    $this->fileStorage = $this->entityTypeManager->getStorage('file');
  }

  /**
   * Loop through all fields an find those
   * who contain a reference to a file entity.
   *
   * @param $entity
   *    The current entity.
   *
   * @return array
   *    Array of field names.
   */
  private function getAllFieldsWithFiles($entity): array {
    $fields = [];

    // Find all fields with files.
    $field_definitions = $entity->getFieldDefinitions();
    foreach ($field_definitions as $field_definition) {
      $type = $field_definition->getType();
      if ($type === 'image' || $type === 'file') {
        $fields[$field_definition->getName()] = $field_definition;
      }
    }
    return $fields;
  }

  /**
   * Find every file used of a media entity.
   *
   * This can be many files depending on usage of
   * translations and revisions.
   *
   * @param $entity
   *   The current media entity.
   *
   * @return array
   *   All files used by the entity.
   */
  private function getAllFilesFromEntity($entity): array {
    $files = [];

    // Get all revisions to not overlook an old file.
    $entity_revisions = $this->mediaStorage->getQuery()
      ->allRevisions()
      ->condition($entity->getEntityType()->getKey('id'), $entity->id())
      ->execute();
    $entity_revisions = $this->mediaStorage->loadMultipleRevisions(array_keys($entity_revisions));

    $fields = $this->getAllFieldsWithFiles($entity);

    foreach ($entity_revisions as $revision) {
      array_push($files, ...$this->getAllFilesPerLanguage($revision, $fields));
    }

    return $files;
  }

  /**
   * Convert a uri from private to public.
   *
   * Used as target to file movement.
   *
   * @param $uri
   *   The current uri.
   *
   * @return string
   *   The new uri.
   */
  private function changeFilePathToPrivate($uri): string {
    if (strpos($uri, 'private://') === FALSE) {
      $uri = str_replace('public://', 'private://', $uri);
    }
    return $uri;
  }

  /**
   * Convert a uri from public to private.
   *
   * Used as target to file movement.
   *
   * @param $uri
   *   The current uri.
   *
   * @return string
   *   The new uri.
   */
  private function changeFilePathToPublic($uri): string {
    if (strpos($uri, 'public://') === FALSE) {
      $uri = str_replace('private://', 'public://', $uri);
    }
    return $uri;
  }

  /**
   * Delete or mark file as temporary.
   *
   * @param \Drupal\file\Entity\File $file
   *   The file.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  private function deleteFile(File $file): void {
    if ($this->deleteAction === 'delete') {
      $file->delete();
      \Drupal::logger('media_files_handler')->info(
        'Deleted file ' . implode(', ', $file->id())
      );
    }
    elseif ($this->deleteAction === 'temporary') {
      // Remove all file usages for current file,
      // otherwise cron will not delete the file.
      $file_usage = \Drupal::service('file.usage')->listUsage($file);
      if (!empty($file_usage)) {
        foreach ($file_usage as $module => $usage) {
          \Drupal::service('file.usage')->delete($file, $module);
        }
      }

      $file->setTemporary();
      $file->save();
      \Drupal::logger('media_files_handler')->info(
        'Marked file ' . $file->id() . ' as temporary.'
      );
    }
  }

  /**
   * Loop through all languages used by the current entity
   * and find all used files.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *    The current entity.
   * @param array $fields
   *    The fields with files.
   *
   * @return array
   *    All files found.
   */
  private function getAllFilesPerLanguage(EntityInterface $entity, array $fields): array {
    $files = [];

    // Loop through all languages to not miss translated files.
    $languages = $entity->getTranslationLanguages(TRUE);
    foreach ($languages as $language) {
      $langcode = $language->getId();
      if ($entity->hasTranslation($langcode)) {
        $entity = $entity->getTranslation($langcode);

        foreach ($fields as $field_name => $field) {
          if ($entity->hasField($field_name) && !$entity->get($field_name)->isEmpty()) {
            $values = $entity->get($field_name)->getValue();
            foreach ($values as $value) {
              if ($value['target_id'] !== NULL) {
                $files[$value['target_id']] = $value['target_id'];
              }
            }
          }
        }
      }
    }

    return $files;
  }

  /**
   * Moves a file to private storage, so an access check can be performed.
   *
   * Private files are only accessible if the user
   * has the permission to do so. The permission is most times
   * coupled to the access permission to the parent entity.
   * E.g. if the media entity is unpublished, a user can only access
   * the file, if he has the permission to view the unpublished media entity.
   *
   * @param \Drupal\file\Entity\File $file
   */
  private function moveFileToPrivateStorage(File $file): void {
    $uri = $file->getFileUri();
    $new_uri = $this->changeFilePathToPrivate($uri);
    $directory = $this->fileSystem->dirname($new_uri);

    // Check if the new directory exists.
    if (!file_exists($directory)) {
      $this->fileSystem->mkdir($directory, NULL, TRUE);
    }

    // Move file. This will also delete image styles.
    file_move($file, $new_uri);

    \Drupal::logger('media_files_handler')->info(
      'Moved file ' . $file->id() . ' to private storage.'
    );
  }

  /**
   * Move all files of an entity to private storage.
   *
   * E.g. when a media entity is unpublished.
   *
   * @param $entity
   *   The current entity.
   */
  public function moveAllFilesToPrivate($entity) :void {
    $files_to_move = $this->getAllFilesFromEntity($entity);

    /** @var \Drupal\file\Entity\File[] $files */
    $files = $this->fileStorage->loadMultiple($files_to_move);
    foreach ($files as $file) {
      $this->moveFileToPrivateStorage($file);
    }
  }

  /**
   * Removes all files used by an entity.
   *
   * E.g. when a media entity is deleted.
   *
   * @param $entity
   *    The current entity.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function removeAllFilesOfEntity($entity) :void {
    $files_to_delete = $this->getAllFilesFromEntity($entity);

    /** @var \Drupal\file\Entity\File[] $files */
    $files = $this->fileStorage->loadMultiple($files_to_delete);
    foreach ($files as $file) {
      $this->deleteFile($file);
    }
  }

  /**
   * Move the files used by the default revision to public storage.
   *
   * E.g. when a media entity is published.
   *
   * @param $entity
   *   The current entity.
   */
  public function moveDefaultRevisionFilesToPublic($entity): void {
    $fields = $this->getAllFieldsWithFiles($entity);

    // Check if fields setting is public, otherwise skip the field. We do not want to
    // accidentally allow access to files which should be checked by access rules.
    foreach ($fields as $field) {
      $item_definition = $field->getItemDefinition();
      if ($item_definition->getSetting('uri_scheme') !== 'public') {
        unset($field, $fields);
      }
    }

    // Proceed if there are any fields with file storage left.
    if (isset($fields) && is_array($fields)) {
      // todo: make sure we have the correct language and revision loaded.
      // todo: make sure files from published translations are used.
      $files_to_move = $this->getAllFilesPerLanguage($entity, $fields);

      // Move the files to public.
      /** @var \Drupal\file\Entity\File[] $files */
      $files = $this->fileStorage->loadMultiple($files_to_move);
      foreach ($files as $file) {
        $uri = $file->getFileUri();

        // If the file is already public, do nothing.
        if (strpos($uri, 'public://') === 0) {
          continue;
        }

        $new_uri = $this->changeFilePathToPublic($uri);
        $directory = $this->fileSystem->dirname($new_uri);

        // Check if the new directory exists.
        if (!file_exists($directory)) {
          $this->fileSystem->mkdir($directory, NULL, TRUE);
        }

        // Move file.
        file_move($file, $new_uri);

        \Drupal::logger('media_files_handler')->info(
          'Moved file ' . $file->id() . ' to public storage.'
        );
      }
    }
  }

  /**
   * Handle the replacement of files in an entity.
   *
   * @param $entity
   *   The current entity.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function handleNewFiles($entity): void {
    // Check if the entity has a new file.
    // If true, delete the old one, if not used by any other language or revision.
    $fields = $this->getAllFieldsWithFiles($entity);
    $files = $this->getAllFilesFromEntity($entity);
    $files_of_current_revision = $this->getAllFilesPerLanguage($entity, $fields);

    // Get files from old entity.
    $old_files = [];
    foreach ($fields as $field_name => $field) {
      $original_entity = $entity->original;
      if ($original_entity !== NULL && $original_entity->hasField($field_name)) {
        $values = $original_entity->get($field_name)->getValue();
        foreach ($values as $value) {
          $old_files[$value['target_id']] = $value['target_id'];
        }
      }
    }

    // Check if old files are used somewhere else. If not, delete.
    foreach ($old_files as $old_file) {
      if (!in_array($old_file, $files, TRUE)) {
        /** @var \Drupal\file\Entity\File $file */
        $file = $this->fileStorage->load($old_file);
        if ($file !== NULL) {
          $this->deleteFile($file);
        }
      }
      elseif (!in_array($old_file, $files_of_current_revision, TRUE)) {
        /** @var \Drupal\file\Entity\File $file */
        $file = $this->fileStorage->load($old_file);
        if ($file !== NULL) {
          $this->moveFileToPrivateStorage($file);
        }
      }
    }
  }

  /**
   * Remove files of specific revision.
   *
   * @param $entity
   *   The current entity.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function removeFilesOfEntityRevision($entity) :void {
    // todo: finish functionality
    $fields = $this->getAllFieldsWithFiles($entity);
    $files = $this->getAllFilesFromEntity($entity);

    $files_from_revision = [];
    foreach ($fields as $field_name => $field) {
      if ($entity->original->hasField($field_name)) {
        $values = $entity->original->get($field_name)->getValue();
        foreach ($values as $value) {
          $files_from_revision[$value['target_id']] = $value['target_id'];
        }
      }
    }

    // Check if old revision files are used somewhere else. If not, delete.
    foreach ($files_from_revision as $revision_file) {
      if (!in_array($revision_file, $files, TRUE)) {
        $this->deleteFile($revision_file);
      }
    }
  }
}
